/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.pdf.textsegmentation.implementation;

import com.frojasg1.chesspdfbrowser.configuration.ChessStringsConf;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.ChessViewConfiguration;
import com.frojasg1.chesspdfbrowser.pdf.textsegmentation.PixelComponents;
import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Usuario
 */
public class SummarizedPage
{
	protected static final double PERCENTAGE_OF_POINTS_TO_MATCH_A_BACKGROUND_PIXEL = 75D;
//	protected static final double PERCENTAGE_OF_POINTS_TO_MATCH_A_VERTICAL_GAP = 85D;
	protected static final int MINIMUM_NUMBER_OF_PAGES_TO_VALIDATE = 3;

	protected static final int THRESHOLD_FOR_INK_COMPONENT = 64;

	protected Dimension _size = null;
	protected int _minimumNumberOfPagesToSummarize = 3;
	protected int _maximumNumberOfPagesToSummarize = 3;

//	protected Map< Integer, SummarizedHomogeneousPoint[] > _summarizedGapPoints_map = null;
	protected PixelComponents[][] _summarizedBackgroundPoints = null;

	protected List< int[] > _listOfPixelsOfPages = null;

	public SummarizedPage( )
	{
		_listOfPixelsOfPages = new ArrayList< int[] >();
	}

	public PixelComponents[][] getSummarizedPixels()
	{
		return( _summarizedBackgroundPoints );
	}

	public Dimension getSize()
	{
		return( _size );
	}

	public void setNumberOfPagesToSummarize( int minimumNumberOfPagesToSummarize,
												int maximumNumberOfPagesToSummarize )
	{
		_minimumNumberOfPagesToSummarize = minimumNumberOfPagesToSummarize;
		_maximumNumberOfPagesToSummarize = maximumNumberOfPagesToSummarize;
	}

	public boolean isComplete()
	{
		return( _summarizedBackgroundPoints != null );
	}

	public void addPage( BufferedImage image )
	{
		if( _listOfPixelsOfPages != null )
		{
			if( _size == null )
			{
				_size = new Dimension( image.getWidth(), image.getHeight() );
			}

			if( ( image.getWidth() != _size.getWidth() ) ||
				( image.getHeight() != _size.getHeight() ) )
			{
				throw( new RuntimeException( getChessStrConf().getProperty( ChessStringsConf.CONF_ERROR_SIZE_NOT_EXPECTED ) +
												" : [" + image.getWidth() + "," + image.getHeight() + "] <> [" +
												+ _size.getWidth() + "," + _size.getHeight() + "]" ) );
			}

			_listOfPixelsOfPages.add( getPixels( image ) );

			if( _listOfPixelsOfPages.size() >= _maximumNumberOfPagesToSummarize )
			{
				summarize();
			}
		}
	}

	protected int[] getPixels( BufferedImage image )
	{
		ImageSegmentator is = new ImageSegmentator( null, image, null );
		return( is.getPixels() );
	}

	public void summarize()
	{
		if( ( _listOfPixelsOfPages != null ) && ( _listOfPixelsOfPages.size() >= _minimumNumberOfPagesToSummarize ) )
		{
			Object[] arrayOfPages = _listOfPixelsOfPages.toArray();

			_summarizedBackgroundPoints = new PixelComponents[ (int) _size.getWidth() ][ (int) _size.getHeight() ];

			int offset = 0;
			for( int xx=0; xx<(int) _size.getWidth(); xx++ )
			{
				offset = xx;

				for( int yy=0; yy<(int) _size.getHeight(); yy++ )
				{
					short[][] arrayOfValuesOfPixels = getArrayOfValuesOfPixels( arrayOfPages, offset );

					PixelComponents shp = ImageSegmentator.getSummararizedHomogeneousPoint(arrayOfValuesOfPixels,
																		PERCENTAGE_OF_POINTS_TO_MATCH_A_BACKGROUND_PIXEL,
																		ImageSegmentator.TOLERANCE_TO_MATCH_COMPONENT_OF_PIXEL_FOR_BACKGROUND_GAP );
					
					_summarizedBackgroundPoints[xx][yy] = shp;
					offset += (int) _size.getWidth();
				}
			}

			_listOfPixelsOfPages = null;
		}
	}

	public short[][] getArrayOfValuesOfPixels( Object[] arrayOfPages, int offset )
	{
		short[][] result = new short[arrayOfPages.length][];

		for( int ii=0; ii<arrayOfPages.length; ii++ )
		{
			int[] arrayOfPixels = (int[]) arrayOfPages[ii];

			int pixel = arrayOfPixels[offset];
			result[ii] = ImageSegmentator.getPixelComponents(pixel);
		}

		return( result );
	}

	protected static ChessStringsConf getChessStrConf()
	{
		return( ChessStringsConf.instance() );
	}
}
