/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.io.writers.implementation;

import com.frojasg1.chesspdfbrowser.exception.ChessParserException;
import com.frojasg1.chesspdfbrowser.io.notation.ChessMoveAlgebraicNotation;
import com.frojasg1.chesspdfbrowser.io.notation.ChessMoveNotation;
import com.frojasg1.chesspdfbrowser.io.writers.ChessGamePGNWriterObserver;
import com.frojasg1.chesspdfbrowser.model.chess.board.ChessBoard;
import com.frojasg1.chesspdfbrowser.model.chess.configuration.ChessLanguageConfiguration;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGame;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGameHeaderInfo;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeNode;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.NAG;
import com.frojasg1.general.StringFunctions;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 *
 * @author Usuario
 */
public class ChessGamePGNwriter extends ChessGameViewWriter
{
	protected static int ADVISED_MAX_LINE_LENGTH = 80;

	protected ChessGamePGNWriterObserver _observer = null;
//	protected ChessMoveNotation _chessMoveNotation = null;

	protected StringBuilder _currentLine = null;

	public ChessGamePGNwriter( ChessGamePGNWriterObserver observer )
	{
		super( false );	// isToWriteInTextPane == false

		_observer = observer;
		_currentLine = new StringBuilder();
	}

	protected void addTagToStringBuilder( String tagKey, String tagValue )
	{
		addStringToStringBuilderMoveTree( "[" + tagKey + " \"" );
		addStringToStringBuilderMoveTree( tagValue );
		addStringToStringBuilderMoveTree( "\"]" );
		processCurrentLine( true );
	}

	@Override
	public String getGameHeaderString( ChessGame chessGame )
	{
		_currentLine = new StringBuilder();
		for( int ii=0; ii<ChessGameHeaderInfo._mandatoryTagArray.length; ii++ )
		{
			String tagKey = ChessGameHeaderInfo._mandatoryTagArray[ii];
			String tagValue = chessGame.getChessGameHeaderInfo().get( tagKey );
			addTagToStringBuilder( tagKey, tagValue );
		}

		for( int jj=0; jj<ChessGameHeaderInfo._optionalTagArrays.length; jj++ )
		{
			String[] groupOfTags = ChessGameHeaderInfo._optionalTagArrays[jj];
			for( int ii=0; ii<groupOfTags.length; ii++ )
			{
				String tagKey = groupOfTags[ii];
				String tagValue = chessGame.getChessGameHeaderInfo().get( tagKey );
				if( tagValue != null )
					addTagToStringBuilder( tagKey, tagValue );
			}
		}

		String fenString = chessGame.getChessGameHeaderInfo().get( ChessGameHeaderInfo.FEN_TAG );
		if( ( fenString != null ) && ( fenString.length() >0 ) )
			addTagToStringBuilder( ChessGameHeaderInfo.FEN_TAG, fenString );
		
		return( null );
	}

	@Override
	protected void addMoveToStringBuilderMoveTree( ChessBoard chessBoard,
											MoveTreeNode mt, MoveTreeNode parent, boolean isMainLine,
											boolean hasToWriteEllipsis ) throws ChessParserException
	{
		String numberOfMove = getNumberOfMoveString(mt.getLevel(), hasToWriteEllipsis );

		addStringToStringBuilderMoveTree( numberOfMove );

		String moveString = "void";
/*
		if( ( mt.getAdditionalInfo() == null ) ||
			( !mt.getAdditionalInfo().getIsIllegalMove() &&
				!mt.getAdditionalInfo().getIsAnyParentIllegalMove() )
		  )
		{
			moveString = ChessMoveAlgebraicNotation.getInstance().getMoveString( chessBoard, mt.getMove() );
		}
		else
*/
		if( mt.getMove().getMoveToken() != null )
		{
			moveString = mt.getMove().getMoveToken().getString();
		}

		addStringToStringBuilderMoveTree( moveString );

		if( ( mt.getMove() != null ) &&
			( mt.getMove().getMoveToken() != null ) )
		{
			Iterator<NAG> it = mt.getMove().getMoveToken().nagIterator();
			while( it.hasNext() )
			{
				NAG nag = it.next();
				if( !_isToWriteInTextPane )
					addStringToStringBuilderMoveTree( " " + nag.getStringForPGNfile() );
				else
					addStringToStringBuilderMoveTree( nag.getStringToShow() );
			}
		}

		if( mt.getComment() != null )
		{
			addStringToStringBuilderMoveTree( " {" );
			addStringToStringBuilderMoveTree( mt.getComment() );
			addStringToStringBuilderMoveTree( "}" );
		}

		if( ( mt.getMove() != null ) && ( mt.getMove()._resultOfGame != null ) )
		{
			addStringToStringBuilderMoveTree( " " + mt.getMove()._resultOfGame );
		}
	}

	@Override
	protected void addStringToStringBuilderMoveTree( String str )
	{
		if( ( str != null ) && ( str.length() != 0 ) )
		{
			_currentLine.append( str );
			processCurrentLine( false );
		}
	}

	@Override
	protected void endOfGame()
	{
		processCurrentLine( true );
	}

	protected void processCurrentLine( boolean forced )
	{
		if( forced || ( _currentLine.length() > ADVISED_MAX_LINE_LENGTH ) )
		{
			if( _currentLine.length() > ADVISED_MAX_LINE_LENGTH )
			{
				Object[] splitted = splitCurrentLine( _currentLine );
				for( int ii=0; ii<(splitted.length-1); ii++ )
				{
					addLineToFile( (String) splitted[ii] );
				}
				_currentLine.append( splitted[ splitted.length-1 ] );
			}
			else
			{
				addLineToFile( _currentLine.toString() );
			}
		}
	}

	protected void addLineToFile( String line )
	{
		_stringBuilderMoveTree.append( line + "\n" );
		if( _observer != null )
			_observer.writeLine_pgnWriter( line.toString() );

		_currentLine = new StringBuilder();
	}
	
	protected Object[] splitCurrentLine( StringBuilder sb )
	{
		List<String> splittedList = new ArrayList<String>();
		int pos = 0;
		boolean continueProcessing = true;
		while( continueProcessing )
		{
			if( sb.length() - pos > ADVISED_MAX_LINE_LENGTH )
			{
				int pos1 = pos + ADVISED_MAX_LINE_LENGTH - 1;
				pos1 = StringFunctions.lastIndexOfAnyChar(sb, " {}", pos1 );
				if( pos1 == -1 )
				{
					pos1 = pos + ADVISED_MAX_LINE_LENGTH;
					pos1 = StringFunctions.indexOfAnyChar(sb, " {}", pos1 );
					if( pos1 == -1 )
					{
						pos1 = sb.length() - 1;
					}
				}
				splittedList.add( sb.substring( pos, pos1 + 1 ) );
				pos = pos1 + 1;
				continueProcessing = ( pos < sb.length() );
			}
			else
			{
				splittedList.add( sb.substring( pos, sb.length() ) );
				continueProcessing = false;
			}
		}

		return( splittedList.toArray() );
	}
}
