/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.chesspdfbrowser.view.chess;

import com.frojasg1.chesspdfbrowser.configuration.ApplicationConfiguration;
import com.frojasg1.chesspdfbrowser.io.writers.WrittenNodeInfo;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGame;
import com.frojasg1.chesspdfbrowser.model.chess.game.ChessGameTreeAdditionalInfo;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeGame;
import com.frojasg1.chesspdfbrowser.model.chess.game.tree.MoveTreeNode;
import com.frojasg1.chesspdfbrowser.view.chess.editcomment.EditCommentFrame;
import com.frojasg1.chesspdfbrowser.view.chess.interaction.ChessGameControllerInterface;
import com.frojasg1.chesspdfbrowser.view.chess.interaction.ChessGameViewInterface;
import com.frojasg1.chesspdfbrowser.view.chess.multiwindowmanager.MultiwindowGameManager;
import com.frojasg1.applications.common.configuration.LanguageConfiguration;
import com.frojasg1.general.desktop.dialogs.DesktopViewComponent;
import com.frojasg1.general.desktop.mouse.AvailableCursors;
import com.frojasg1.general.desktop.view.SelectionPreservingCaret;
import com.frojasg1.general.desktop.view.menus.BaseJPopupMenu;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.generic.GenericFunctions;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.swing.JEditorPane;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPopupMenu;
import javax.swing.JSeparator;
import javax.swing.JTextPane;
import javax.swing.SwingUtilities;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

/**
 *
 * @author Usuario
 */
public class ChessTreeGameTextPane extends JTextPane implements MouseMotionListener, MouseListener, LanguageConfiguration, DesktopViewComponent
{
	protected static final String CONF_COULD_NOT_COPY_ALREADY_A_MOVE = "COULD_NOT_COPY_ALREADY_A_MOVE";
	protected static final String CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT = "PROBLEM_WHILE_COPYING_A_SUBVARIANT";
	protected static final String CONF_COULD_NOT_COPY_NUMBER_OF_PLY_DOES_NOT_MATCH = "COULD_NOT_COPY_NUMBER_OF_PLY_DOES_NOT_MATCH";
	protected static final String CONF_COULD_NOT_COPY_NO_SUBVARIANT = "COULD_NOT_COPY_NO_SUBVARIANT";
	protected static final String CONF_NODE_TO_PASTE_TO_THE_COPIED_SUBVARIANT_NOT_SELECTED = "NODE_TO_PASTE_TO_THE_COPIED_SUBVARIANT_NOT_SELECTED";

	protected static final Cursor sa_handCursor = new Cursor( Cursor.HAND_CURSOR );
	protected static final Cursor sa_defaultCursor = new Cursor( Cursor.DEFAULT_CURSOR );

	// The next values are for background type:
	protected static final int _NORMAL = 0;				// white background
	protected static final int _SELECTED = 1;			// yellow background
	protected static final int _COPY_SELECTION = 2;		// grey background
	protected static final int _ILLEGAL_MOVE = 3;		// red background
	protected static final int _NUMBER_OF_BT = 4;

	// The next values are for the style names for moves.
	protected static final int _PLAIN = 0;		// illegal or undefined move
	protected static final int _BOLD_BLACK = 1;	// main variant
	protected static final int _BLUE = 2;		// secondary variant
	protected static final int _GREEN = 3;		// comment
	protected static final int _NUMBER_OF_STYLE_NAMES = 4;

	protected Style[][] _styles = null;


	protected final String BOLD_BLACK = "ChessTreeGameTextPane_BOLD_BLACK";
//	protected final String BOLD_BLACK_SELECTED = "ChessTreeGameTextPane_BOLD_BLACK_SELECTED";
	protected final String GREEN = "ChessTreeGameTextPane_GREEN";
	protected final String BLUE = "ChessTreeGameTextPane_BLUE";
//	protected final String BLUE_SELECTED = "ChessTreeGameTextPane_BLUE_SELECTED";
	protected final String PLAIN = "ChessTreeGameTextPane_PLAIN";
	protected final String LOCAL_DEFAULT_STYLE = "ChessTreeGameTextPane_DEFAULT_STYLE";

	protected final String SELECTED_SUFFIX = "_SELECTED";
	protected final String ILLEGAL_MOVE_SUFFIX = "_ILLEGAL_MOVE";
	protected final String COPY_SELECTION_SUFFIX = "_COPY_SELECTION";

	protected MultiwindowGameManager _parent;

	protected ChessGame _chessGame = null;

	protected ChessGameControllerInterface _controller = null;
	protected ChessGameViewInterface _chessBoardPanel = null;

	protected WrittenNodeInfo _previousSelectedMoveNodeInfo = null;

//	protected MoveTreeNode _previousVariantSelectionMoveTreeNode = null;
//	protected MoveTreeNode _variantSelectionMoveTreeNode = null;

	protected MoveTreeNode _lastClickMoveTreeNodeForPopupMenu = null;
	protected MoveTreeNode _previousSelectedMove = null;

	protected ContextualMenu _popupMenu = null;

	protected boolean _browseGameEnabled = false;

	protected LanguageConfiguration _languageConfiguration = null;
/*
	protected static enum BackgroundType
	{
		NORMAL,
		SELECTED,
		COPY_SELECTION,
		ILLEGAL_MOVE
	}
*/

	public ChessTreeGameTextPane( ChessGameViewInterface chessBoardPanel,
									ChessGameControllerInterface controller,
									MultiwindowGameManager parent,
									boolean browseGameEnabled )
	{
		_chessBoardPanel = chessBoardPanel;
		_controller = controller;

		setEditable( false );
		setCaret( new SelectionPreservingCaret() );
		
		int defaultFontSize = -1;
		addStyles( defaultFontSize );

		createStyleArrayCached();
		
//		System.out.println( "ChessTreeGameTextPane, addMouseListener" );
		addMouseListener( this );
		addMouseMotionListener( this );

		_popupMenu = new ContextualMenu();

		_parent = parent;

		_browseGameEnabled = browseGameEnabled;
	}

	public ChessGameControllerInterface getChessGameController()
	{
		return( _controller );
	}
	
	public JPopupMenu getJPopupMenu()
	{
		return( _popupMenu );
	}

	public void setChessGame( ChessGame chessGame )
	{
		_chessGame = chessGame;
	}

	public ChessGame getChessGame()
	{
		return( _chessGame );
	}
	
	public MoveTreeNode getPreviousVariantSelectionMoveTreeNode()
	{
		return( _parent.getPreviousVariantSelectionMoveTreeNode() );
	}

	public MoveTreeNode getVariantSelectionMoveTreeNode()
	{
		return( _parent.getVariantSelectionMoveTreeNode() );
	}
/*
	public void setPreviousVariantSelectionMoveTreeNode( MoveTreeNode node )
	{
		_parent.setPreviousVariantSelectionMoveTreeNode(node);
	}
*/
	public void setVariantSelectionMoveTreeNode( MoveTreeNode node )
	{
		_parent.setVariantSelectionMoveTreeNode(node);
	}

	public void update( boolean everyThing )
	{
		if( ! everyThing )
			updateViewPositionInMovesTree();
		else
			updateAfterinsertionOfMovesInTree();
	}

	protected void changeSelectedMove( WrittenNodeInfo nodeInfo )
	{
		try
		{
	//		System.out.println( "Actualizando lista de movimientos." );
			_chessGame.setCurrentListOfMoves( nodeInfo.getMoveTreeNode().getGameMoveList() );
	//		updateAfterinsertionOfMovesInTree();
			updateViewPositionInMovesTree();
	//		System.out.println( "actualizada insercion de movimientos en arbol." );
			EditCommentFrame.instance().setMove( _chessGame, _chessGame.getCurrentMove() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	@Override
	public void mouseClicked(MouseEvent evt)
	{
//		System.out.println( "mouseClicked." );
		JEditorPane editor = (JEditorPane) evt.getSource();
		Point pt = new Point(evt.getX(), evt.getY());
		int pos = editor.viewToModel(pt);

		ChessGameTreeAdditionalInfo cmtai = _chessGame.getChessGameTreeAdditionalInfo();

		if( cmtai != null )
		{
			WrittenNodeInfo nodeInfo = cmtai.getMoveAdditionalInfo(pos);

			if( SwingUtilities.isLeftMouseButton(evt) )
			{
				if( ( nodeInfo != null ) && nodeInfo.isInsideMove(pos) )
				{
					if( _browseGameEnabled &&
						( _chessGame != null ) ) //&& _chessGame.isValid() &&
//						!nodeInfo.getIsIllegalMove() &&
//						!nodeInfo.getIsAnyParentIllegalMove() )	// if chessGame do not have initial position, we do not let to go to the movement
					{
						changeSelectedMove( nodeInfo );
					}
					else
					{
						doPopup( evt, nodeInfo );
					}
				}
				else if( ( nodeInfo != null ) && nodeInfo.isInsideComment(pos) )
				{
					editComment( nodeInfo );
				}
			}
			else if( SwingUtilities.isRightMouseButton(evt) )
			{
				doPopup( evt, nodeInfo );
			}
		}
		else
		{
			if( SwingUtilities.isRightMouseButton(evt) )
			{
				doPopup( evt, null );
			}
		}
	}

	@Override
	public void mouseEntered(MouseEvent e)
	{
		
	}

	@Override
	public void mouseExited(MouseEvent e)
	{
		
	}

	@Override
	public void mousePressed(MouseEvent e)
	{
		
	}

	@Override
	public void mouseReleased(MouseEvent e)
	{
		
	}

	@Override
	public void mouseDragged(MouseEvent e)
	{
		
	}

	@Override
	public void mouseMoved(MouseEvent evt)
	{
		if( _browseGameEnabled &&
			( _chessGame != null ) &&
			_chessGame.isValid() &&			// if the chessGame does not have intial position, we dont show the hand.
			! getCursor().getName().equals( AvailableCursors._waitCursor.getName() ) )	// if we are loading or saving, we must not set the hand cursor.
		{
			JEditorPane editor = (JEditorPane) evt.getSource();
			Point pt = new Point(evt.getX(), evt.getY());
			int pos = editor.viewToModel(pt);

			ChessGameTreeAdditionalInfo cmtai = _chessGame.getChessGameTreeAdditionalInfo();

			if( cmtai != null )
			{
				WrittenNodeInfo nodeInfo = cmtai.getMoveAdditionalInfo(pos);
				if( ( nodeInfo == null ) ||
						!nodeInfo.isInsideMove(pos) &&
						!nodeInfo.isInsideComment(pos) ||
						nodeInfo.getIsIllegalMove() ||
						nodeInfo.getIsAnyParentIllegalMove() )
				{
					setCursor(sa_defaultCursor);
				}
				else
				{
					setCursor( sa_handCursor );
				}
			}
		}
	}

	public void updateViewPositionInMovesTree()
	{
		if( _chessGame != null )
		{
			if( ( getPreviousVariantSelectionMoveTreeNode() != null ) &&
				( getPreviousVariantSelectionMoveTreeNode().getRootNode() == _chessGame.getMoveTreeGame() ) )
			{
				formatCompleteVariant(_NORMAL, getPreviousVariantSelectionMoveTreeNode() );
			}

			if( ( getVariantSelectionMoveTreeNode() != null ) &&
				( getVariantSelectionMoveTreeNode().getRootNode() == _chessGame.getMoveTreeGame() ) )
			{
				formatCompleteVariant( _COPY_SELECTION, getVariantSelectionMoveTreeNode() );
			}

//			setPreviousVariantSelectionMoveTreeNode(null);

			if( _browseGameEnabled )
			{
				if( ( _previousSelectedMoveNodeInfo != null ) &&
					( _previousSelectedMoveNodeInfo.getMoveTreeNode() != getVariantSelectionMoveTreeNode() ) &&
					!_previousSelectedMoveNodeInfo.getMoveTreeNode().isParent(getVariantSelectionMoveTreeNode()) )
				{
					formatMoveText( _NORMAL, _previousSelectedMoveNodeInfo, false );
				}

				WrittenNodeInfo nodeInfo = _chessGame.getCurrentMove().getAdditionalInfo();

				if( _previousSelectedMove != null )
				{
					if( _previousSelectedMove.getAdditionalInfo().getIsIllegalMove() )
						formatMoveText( _ILLEGAL_MOVE, _previousSelectedMove.getAdditionalInfo(), false );
					else if( ( _previousSelectedMove == getVariantSelectionMoveTreeNode() ) ||
							( _previousSelectedMove.isParent( getVariantSelectionMoveTreeNode() ) ) )
					{
						formatMoveText( _COPY_SELECTION, _previousSelectedMove.getAdditionalInfo(), false );
					}
					else
					{
						formatMoveText( _NORMAL, _previousSelectedMove.getAdditionalInfo(), false );
					}
				}

				if( nodeInfo != null )
				{
					requestFocus();
					setSelectionStart( nodeInfo.getInitialPositionOfMove() );
					setSelectionEnd( nodeInfo.getInitialPositionOfMove() );
					formatMoveText( _SELECTED, nodeInfo, false );
					_previousSelectedMove = _chessGame.getCurrentMove();
				}
				else
				{
					requestFocus();
					setSelectionStart( 0 );
					setSelectionEnd( 0 );
				}

				_previousSelectedMoveNodeInfo = nodeInfo;
			}

			repaint();
			if( _chessBoardPanel != null )
			{
				_chessBoardPanel.refresh();
			}
		}
	}

	protected void formatCompleteVariant( int bt, MoveTreeNode mtn )
	{
		WrittenNodeInfo wni = mtn.getAdditionalInfo();
		formatMoveText( bt, wni, false );

		for( int ii=0; ii<mtn.getNumberOfChildren(); ii++ )
			formatCompleteVariant( bt, mtn.getChild( ii ) );
	}

	protected String getSuffixForStyleNameFromBackgroundType( int bt )
	{
		String result = "";
		switch( bt )
		{
			case _NORMAL:			result = ""; break;
			case _SELECTED:			result = SELECTED_SUFFIX; break;
			case _COPY_SELECTION:	result = COPY_SELECTION_SUFFIX; break;
			case _ILLEGAL_MOVE:		result = ILLEGAL_MOVE_SUFFIX; break;
			default:
				result = "";
		}
		return( result );
	}

	protected String getStyleName( int styleName )
	{
		String result = "";
		switch( styleName )
		{
			case _PLAIN:				result = PLAIN; break;
			case _BOLD_BLACK:			result = BOLD_BLACK; break;
			case _BLUE:					result = BLUE; break;
			case _GREEN:				result = GREEN; break;
			default:
				result = "";
		}
		return( result );
	}

	protected Style getStyle( int styleName, int bt )
	{
		Style result = null;

		String styleNameStr = getStyleName( styleName );
		String suffixForStyleName = getSuffixForStyleNameFromBackgroundType( bt );

		result = getStyle( styleNameStr + suffixForStyleName );
		
		return( result );
	}

	protected Style getStyleCached( int styleName, int bt )
	{
		return( _styles[styleName][bt] );
	}

	protected void createStyleArrayCached()
	{
		_styles = new Style[_NUMBER_OF_STYLE_NAMES][_NUMBER_OF_BT];

		for( int styleName = 0; styleName < _NUMBER_OF_STYLE_NAMES; styleName++ )
			for( int bt = 0; bt < _NUMBER_OF_BT; bt++ )
			{
				_styles[styleName][bt] = getStyle( styleName, bt );
			}
	}

	protected void formatMoveText( int bt, WrittenNodeInfo wni, boolean hasToFormatComment )
	{
		if( wni.getIsIllegalMove()  )
			bt = _ILLEGAL_MOVE;

		int styleName = -1;
		Style style = null;

		if( //!_chessGame.isValid() ||
			wni.getIsAnyParentIllegalMove() ||
			wni.getIsIllegalMove() )
		{
			styleName = _PLAIN;
		}
		else if( wni.isMainLine() )										styleName = _BOLD_BLACK;
		else if( !wni.isMainLine() )									styleName = _BLUE;

		style = getStyleCached( styleName, bt );

		getStyledDocument().setCharacterAttributes(wni.getInitialPositionOfMove(),
													wni.getFinalPositionOfMove() - wni.getInitialPositionOfMove(),
													style, true);
		
		if( hasToFormatComment )
		{
			if( wni.getInitialPositionOfComment() > 0 )
			{
				style = getStyleCached( _GREEN, bt );

				getStyledDocument().setCharacterAttributes(wni.getInitialPositionOfComment(),
															wni.getFinalPositionOfComment() - wni.getInitialPositionOfComment() + 1,
															style, true);
			}
		}
	}

	protected void formatChessGameMoves( String newString )
	{
		_previousSelectedMoveNodeInfo = null;
		setText( newString );

		getStyledDocument().setCharacterAttributes(0, getStyledDocument().getLength(), getStyle( PLAIN ), true);

		ChessGameTreeAdditionalInfo cgtai = _chessGame.getChessGameTreeAdditionalInfo();

		Style green = getStyle( GREEN );
		Style green_copySelection = getStyle( GREEN + COPY_SELECTION_SUFFIX );

		for( int ii=0; ii<cgtai.size(); ii++ )
		{
			WrittenNodeInfo wni = cgtai.getElement(ii);

			int bt = -1;

			if( wni.getIsIllegalMove() )
			{
				bt = _ILLEGAL_MOVE;
			}
			else
			{
				bt = _NORMAL;
			}

//			if( _chessGame.isValid() ) // if we have intial position, we format the moves. If we do not have initial position we do not format, to distinguish
			{
				formatMoveText( bt, wni, false );
			}

			if( wni.getInitialPositionOfComment() > 0 )
			{
				getStyledDocument().setCharacterAttributes(wni.getInitialPositionOfComment(),
															wni.getFinalPositionOfComment() - wni.getInitialPositionOfComment() + 1,
															green, true);
			}
		}
	}

	protected void updateAfterinsertionOfMovesInTree()
	{
		if( _chessGame != null )
		{
			int selectionStart = getSelectionStart();
			int selectionEnd = getSelectionEnd();
			
			String newString = _chessGame.getChessGameTreeAdditionalInfo().getMoveTreeString();
//			if( ! getText().equals( newString ) )
			{
				formatChessGameMoves( newString );
			}
			updateViewPositionInMovesTree();

			if( ! _browseGameEnabled )
			{
				setSelectionStart( IntegerFunctions.min( selectionStart, getText().length() - 1 ) );
				setSelectionEnd( IntegerFunctions.min( selectionEnd, getText().length() ) );
			}
		}
	}

	public void updateEverythingFully()
	{
		if( _chessGame != null )
		{
			int selectionStart = getSelectionStart();
			int selectionEnd = getSelectionEnd();
			
			String newString = "";

			try
			{
				newString = _chessGame.getChessGameTreeAdditionalInfo().updateAdditionalInfo();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}

//			if( ! getText().equals( newString ) )
			{
				formatChessGameMoves( newString );
			}
			updateViewPositionInMovesTree();

			if( ! _browseGameEnabled )
			{
				setSelectionStart( IntegerFunctions.min( selectionStart, getText().length() - 1 ) );
				setSelectionEnd( IntegerFunctions.min( selectionEnd, getText().length() ) );
			}
		}
	}

	protected void addStyles( int defaultFontSize )
	{
		StyledDocument sd = getStyledDocument();

		Style defaultStyle = sd.getStyle(StyleContext.DEFAULT_STYLE);

		Style defaultStyleFontSize = sd.addStyle( LOCAL_DEFAULT_STYLE, defaultStyle);
		if( defaultFontSize > 0 )
			StyleConstants.setFontSize(defaultStyleFontSize, defaultFontSize );

		// Create and add the main document style
		final Style bold = sd.addStyle(BOLD_BLACK, defaultStyle);
		StyleConstants.setFontSize(bold, StyleConstants.getFontSize( defaultStyle ) );
		StyleConstants.setForeground(bold, Color.BLACK );
		StyleConstants.setBold(bold, true);

		final Style green = sd.addStyle(GREEN, defaultStyle);
		StyleConstants.setForeground(green, Color.green.darker() );
		StyleConstants.setBold(green, false);

		final Style blue = sd.addStyle(BLUE, defaultStyle);
		StyleConstants.setForeground(blue, Color.BLUE.darker() );
		StyleConstants.setBold(blue, false);

		final Style plain = sd.addStyle(PLAIN, defaultStyle);
		StyleConstants.setForeground(plain, Color.BLACK );
		StyleConstants.setBold(plain, false);

		String[] styleNames= new String[] { BOLD_BLACK, BLUE, GREEN, PLAIN };
		addStyles_background( styleNames, SELECTED_SUFFIX, new Color( 224, 224, 224 ) );		// bright grey
		addStyles_background( styleNames, ILLEGAL_MOVE_SUFFIX, new Color( 253, 109, 90 ) );		// red
		addStyles_background( styleNames, COPY_SELECTION_SUFFIX, new Color( 253, 250, 9 ) );	// yellow
	}

	protected void addStyles_background( String[] styleNames, String suffix, Color backGroundColor )
	{
		StyledDocument sd = getStyledDocument();

		for( int ii=0; ii<styleNames.length; ii++ )
		{
			Style baseStyle = sd.getStyle( styleNames[ii] );

			String newStyleName = styleNames[ii] + suffix;
			
			final Style newStyle = sd.addStyle(newStyleName, baseStyle );
			StyleConstants.setBackground(newStyle, backGroundColor );
		}
	}

	protected boolean gameDoNotHaveInitialPosition()
	{
		return( !_chessGame.isValid() );
	}
	
	protected void doPopup( MouseEvent evt, WrittenNodeInfo nodeInfo )
	{
		_lastClickMoveTreeNodeForPopupMenu = null;
		if( nodeInfo != null )
			_lastClickMoveTreeNodeForPopupMenu = nodeInfo.getMoveTreeNode();	
			
		enablePopupMenuItems();

		_popupMenu.show(evt.getComponent(), evt.getX() - 5, evt.getY() - 5);
	}

	protected void enablePopupMenuItems()
	{
		_popupMenu.setAllEnabled(true);
		
		if( _lastClickMoveTreeNodeForPopupMenu == null )
		{
			_popupMenu.setSelectAndCopyEnabled(false);
			_popupMenu.setEraseEnabled(false);
			_popupMenu.setEditCommentEnabled(false);
		}

		if( getVariantSelectionMoveTreeNode() == null )
		{
			_popupMenu.setClearSelectionEnabled(false);
			_popupMenu.setPasteEnabled(false);
		}
	}

	protected void selectAndCopy()
	{
//		setPreviousVariantSelectionMoveTreeNode( getVariantSelectionMoveTreeNode() );
		setVariantSelectionMoveTreeNode( _lastClickMoveTreeNodeForPopupMenu );
		_lastClickMoveTreeNodeForPopupMenu = null;

//		updateViewPositionInMovesTree();
	}

	protected void clearSelection()
	{
//		setPreviousVariantSelectionMoveTreeNode( getVariantSelectionMoveTreeNode() );
		setVariantSelectionMoveTreeNode( null );
		_lastClickMoveTreeNodeForPopupMenu = null;

//		updateViewPositionInMovesTree();
	}

	protected void pasteSubvariant()
	{
		MoveTreeNode nodeToPaste = getVariantSelectionMoveTreeNode();
		if( nodeToPaste != null )
		{
			MoveTreeNode nodeToPasteTo = _lastClickMoveTreeNodeForPopupMenu;
			if( ( _lastClickMoveTreeNodeForPopupMenu == null ) &&
				( _chessGame.getMoveTreeGame().getNumberOfChildren() == 0 ) )
			{
				nodeToPasteTo = _chessGame.getMoveTreeGame();

				_chessGame.setInitialPosition( nodeToPaste.getPreviousPosition() );
			}

			if( nodeToPasteTo != null )
			{
				if( (nodeToPasteTo instanceof MoveTreeGame ) ||
					( nodeToPaste.getLevel() ==
						( nodeToPasteTo.getLevel() + 1 ) ) ) 	// if the plyNumber match
				{
					if ( nodeToPasteTo.findChild( nodeToPaste.getMove() ) == null )	// if the child does not exist
					{
						pasteSubvariant_internal( nodeToPasteTo, nodeToPaste );
/*
						try
						{
							_chessGame.updateAdditionalInfo();
						}
						catch( Throwable th )
						{
							th.printStackTrace();
						}
*/
						_controller.newChessGameChosen( _chessGame, true );
	//					updateAfterinsertionOfMovesInTree();
						boolean everyThing = true;
						_parent.updateChessGameWindows(_chessGame, everyThing );

						_lastClickMoveTreeNodeForPopupMenu = null;
					}
					else
					{
						GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this, getInternationalString( CONF_COULD_NOT_COPY_ALREADY_A_MOVE ) +
																							": " + nodeToPaste.getMove(),
																							getInternationalString( CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT ),
																							DialogsWrapper.ERROR_MESSAGE );
					}
				}
				else
				{
					GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this, String.format( getInternationalString(CONF_COULD_NOT_COPY_NUMBER_OF_PLY_DOES_NOT_MATCH ),
																			_lastClickMoveTreeNodeForPopupMenu.getLevel(),
																			nodeToPaste.getLevel() ),
																			getInternationalString( CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT ),
																			DialogsWrapper.ERROR_MESSAGE );
				}
			}
			else
			{
				GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this, getInternationalString( CONF_NODE_TO_PASTE_TO_THE_COPIED_SUBVARIANT_NOT_SELECTED ),
																			getInternationalString( CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT ),
																			DialogsWrapper.ERROR_MESSAGE );
			}
		}
		else
		{
			GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this, getInternationalString(CONF_COULD_NOT_COPY_NO_SUBVARIANT ),
													getInternationalString( CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT ),
													DialogsWrapper.ERROR_MESSAGE );
		}
	}

	@Override
	public Component getComponent()
	{
		return( this );
	}

	
	protected void pasteSubvariant_internal( MoveTreeNode parent, MoveTreeNode variantToCopy )
	{
		MoveTreeNode mtn = parent.simpleInsert( variantToCopy );

		for( int ii=0; ii<variantToCopy.getNumberOfChildren(); ii++ )
		{
			pasteSubvariant_internal( mtn, variantToCopy.getChild( ii ) );
		}
	}

	protected void erase()
	{
		if( _lastClickMoveTreeNodeForPopupMenu != null )
		{
			MoveTreeNode selectedNode = getVariantSelectionMoveTreeNode();
			
			if( selectedNode != null )
			{
				if( ( selectedNode == _lastClickMoveTreeNodeForPopupMenu ) ||
					( _lastClickMoveTreeNodeForPopupMenu.isParent( selectedNode ) ) ||
					( selectedNode.isParent( _lastClickMoveTreeNodeForPopupMenu ) ) )
				{
					setVariantSelectionMoveTreeNode(null);
					setVariantSelectionMoveTreeNode(null);
				}
			}

			if( ( _chessGame.getCurrentMove() == _lastClickMoveTreeNodeForPopupMenu ) ||
				( _chessGame.getCurrentMove().isParent( _lastClickMoveTreeNodeForPopupMenu ) ) )
			{
				try
				{
					_chessGame.setCurrentMove( _lastClickMoveTreeNodeForPopupMenu.getParent() );
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
			}
			
			_lastClickMoveTreeNodeForPopupMenu.remove();

			_controller.newChessGameChosen( _chessGame, true );
/*
			try
			{
				_chessGame.updateAdditionalInfo();
			}
			catch( Throwable th )
			{
				th.printStackTrace();
			}
*/
//			updateAfterinsertionOfMovesInTree();
			boolean everyThing = true;
			_parent.updateChessGameWindows( _chessGame, everyThing );
		}

		_lastClickMoveTreeNodeForPopupMenu = null;

		updateViewPositionInMovesTree();
	}

	protected void setInitialPosition()
	{
		_controller.editInitialPosition( _chessGame );
	}

	protected void editComment( WrittenNodeInfo nodeInfo )
	{
		if( nodeInfo != null )
		{
			changeSelectedMove( nodeInfo );
			_controller.editComment( _chessGame, nodeInfo.getMoveTreeNode() );
		}
	}

	@Override
	public void changeLanguage(String language) throws Exception
	{
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void registerInternationalString(String label, String value)
	{
		if( _languageConfiguration != null )
			_languageConfiguration.registerInternationalString( label, value );
	}

	@Override
	public String getInternationalString(String label)
	{
		String result = null;

		if( _languageConfiguration != null )
			result = _languageConfiguration.getInternationalString( label );

		return( result );
	}

	public void setLanguageConfiguration( LanguageConfiguration languageConfiguration )
	{
		_languageConfiguration = languageConfiguration;

		registerInternationalString(CONF_COULD_NOT_COPY_ALREADY_A_MOVE, "Could not copy: position to copy the variant, had already a move" );
		registerInternationalString(CONF_PROBLEM_WHILE_COPYING_A_SUBVARIANT, "Problem while copying a subvariant" );
		registerInternationalString(CONF_COULD_NOT_COPY_NUMBER_OF_PLY_DOES_NOT_MATCH, "Could not copy: number of ply does not match: " +
														"Ply number of parent: %d. Ply number of root of variant to copy %d" );
		registerInternationalString(CONF_COULD_NOT_COPY_NO_SUBVARIANT, "Could not copy: There was no subvariant to copy" );
		registerInternationalString(CONF_NODE_TO_PASTE_TO_THE_COPIED_SUBVARIANT_NOT_SELECTED, "Node to paste to the copied subvariant not selected" );
	}

	protected class ContextualMenu extends BaseJPopupMenu
	{
		JMenuItem _menuItem_selectAndCopy = null;
		JMenuItem _menuItem_clearSelection = null;
		JMenuItem _menuItem_paste = null;
		JMenuItem _menuItem_erase = null;
		JMenuItem _menuItem_editComment = null;
		JMenuItem _menuItem_setInitialPosition = null;

		public ContextualMenu()
		{
			_menuItem_selectAndCopy = new JMenuItem( "Select and copy subvariant" );
			_menuItem_selectAndCopy.setName( "_menuItem_selectAndCopy" );

			_menuItem_clearSelection = new JMenuItem( "Clear selection" );
			_menuItem_clearSelection.setName( "_menuItem_clearSelection" );

			_menuItem_paste = new JMenuItem( "Paste" );
			_menuItem_paste.setName( "_menuItem_paste" );

			_menuItem_erase = new JMenuItem( "Erase" );
			_menuItem_erase.setName( "_menuItem_erase" );

			_menuItem_editComment = new JMenuItem( "Edit Comment" );
			_menuItem_editComment.setName( "_menuItem_editComment" );

			_menuItem_setInitialPosition = new JMenuItem( "Edit Initial position of game" );
			_menuItem_setInitialPosition.setName( "_menuItem_setInitialPosition" );

			addMenuComponent( _menuItem_selectAndCopy );
			addMenuComponent( _menuItem_clearSelection );
			addMenuComponent( new JSeparator() );
			addMenuComponent( _menuItem_paste );
			addMenuComponent( _menuItem_erase );
			addMenuComponent( new JSeparator() );
			addMenuComponent( _menuItem_editComment );
			addMenuComponent( new JSeparator() );
			addMenuComponent( _menuItem_setInitialPosition );
			
			addMouseListenerToAllComponents();
		}

		public void setAllEnabled( boolean value )
		{
			_menuItem_selectAndCopy.setEnabled(value);
			_menuItem_clearSelection.setEnabled(value);
			_menuItem_paste.setEnabled(value);
			_menuItem_erase.setEnabled(value);
			_menuItem_editComment.setEnabled(value);
			_menuItem_setInitialPosition.setEnabled(value);
		}

		public void setSelectAndCopyEnabled( boolean value )
		{
			_menuItem_selectAndCopy.setEnabled(value);
		}

		public void setClearSelectionEnabled( boolean value )
		{
			_menuItem_clearSelection.setEnabled(value);
		}

		public void setPasteEnabled( boolean value )
		{
			_menuItem_paste.setEnabled(value);
		}

		public void setEraseEnabled( boolean value )
		{
			_menuItem_erase.setEnabled(value);
		}

		public void setEditCommentEnabled( boolean value )
		{
			_menuItem_editComment.setEnabled(value);
		}

		public void setInitialPositionEnabled( boolean value )
		{
			_menuItem_setInitialPosition.setEnabled(value);
		}

		@Override
		public void actionPerformed( ActionEvent evt )
		{
			Component comp = (Component) evt.getSource();
			
			if( comp == _menuItem_selectAndCopy )
				selectAndCopy();
			if( comp == _menuItem_clearSelection )
				clearSelection();
			else if( comp == _menuItem_paste )
				pasteSubvariant();
			else if( comp == _menuItem_erase )
				erase();
			else if( comp == _menuItem_editComment )
			{
				if( _lastClickMoveTreeNodeForPopupMenu != null )
					editComment( _lastClickMoveTreeNodeForPopupMenu.getAdditionalInfo() );
			}
			else if( comp == _menuItem_setInitialPosition )
				setInitialPosition();

			setVisible(false);
		}

	}
}
