/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.common.configuration;

import com.frojasg1.applications.common.configuration.application.ChangeLanguageClientInterface;
import com.frojasg1.applications.common.configuration.application.ChangeLanguageServerInterface;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.GenericFileFunctions;
import com.frojasg1.general.exceptions.ConfigurationException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

/**
 * 
 *	2016-12-06. In this new version of the class, the subfolders of the final created folder may be compound folders.
 *
 * @author Usuario
 */
public abstract class ConfigurationParent extends Properties
											implements ChangeLanguageClientInterface,
														ChangeLanguageServerInterface
{
	protected ChangeLanguageServerInterface _changeLanguageServer = null;
	protected List<ChangeLanguageClientInterface> _listOfObserversOfLanguageChanges = new ArrayList<ChangeLanguageClientInterface>();
	boolean _areLanguageNotificationsActivated = true;

	protected Properties a_properties = null;
	protected Properties a_defaultProperties = new Properties();
	protected Properties a_defaultDefaultProperties = new Properties();

	protected String a_mainFolder = null;
	protected String a_applicationName = null;
	protected String a_group = null;
	protected String a_language = null;
	protected String a_configurationFileName = null;

	public static String  sa_dirSeparator = System.getProperty( "file.separator" );
	public static String  sa_lineSeparator = System.getProperty( "line.separator" );
	protected String a_rootApplicationConfigurationPath = null;
	protected boolean a_existedConfigurationFile = false;
	protected boolean a_existedConfigurationFileInClassPath = false;

	protected boolean _avoidRecursion = false;
	
//	public static String FROJASG1_FOLDER = "frojasg1.apps";

	protected abstract Properties M_getDefaultProperties( String language );
	
	public ConfigurationParent()
	{
		a_properties = new Properties();
	}
	
	public ConfigurationParent( String mainFolder, String applicationName, String group,
								String language, String configurationFileName )
	{
		a_mainFolder = GenericFileFunctions.convertFolderSeparator( mainFolder );
		a_applicationName = GenericFileFunctions.convertFolderSeparator( applicationName );
		a_group = GenericFileFunctions.convertFolderSeparator( group );
		a_language = language;
		a_configurationFileName = configurationFileName;
		a_properties = new Properties();

		a_rootApplicationConfigurationPath = M_getRootApplicationConfigurationPath();
	}

	protected String M_getRootApplicationConfigurationPath()
	{
		String result = System.getProperty("user.home") + sa_dirSeparator  + 
						a_mainFolder + sa_dirSeparator + a_applicationName;
		if( a_group != null ) result = result + sa_dirSeparator + a_group;
		return( result );
	}
	
	public void M_openConfiguration() throws ConfigurationException
	{
		String fileName = M_getConfigurationFileName( a_language );
		M_openConfiguration( fileName );
	}
	
	public void M_openConfiguration( String fileName ) throws ConfigurationException
	{
		Properties properties = null;
		properties = M_loadConfigurationFile( fileName );
		a_properties = properties; // only if there is no exception, we change the values.

		if( ( a_defaultProperties == null ) || ( a_defaultProperties.size() == 0 ) )
			a_defaultProperties = M_getDefaultProperties(a_language);

		if( (a_properties == null) && (a_defaultProperties == null) )
			throw( new ConfigurationException( "The configuration file " + fileName + " could not be open" ) );
	}

	protected void changeLanguage_internal_simple( String language ) throws ConfigurationException
	{
		try
		{
			if( a_language != null ) M_saveConfiguration();
		}
		catch( ConfigurationException ce )
		{
			ce.printStackTrace();
		}

		try
		{
			a_defaultProperties=M_getDefaultProperties( language );
			M_openConfiguration( M_getConfigurationFileName(language) );
			a_language = language;
		}
		catch( ConfigurationException ce )
		{
			ce.printStackTrace();

			throw( ce );
		}
	}

	@Override
	public void setLanguage( String language )
	{
	}
	
	@Override
	public String getLanguage()
	{
		return( a_language );
	}

	@Override
	public void changeLanguage( String language ) throws ConfigurationException
	{
		boolean revertOnError = true;
		changeLanguage_internal( language, revertOnError );
	}

	protected void changeLanguage_internal( String language, boolean revertOnError ) throws ConfigurationException
	{
		if( !_avoidRecursion )
		{
			String oldLanguage = getLanguage();

			try
			{
				changeLanguage_internal_simple(language);

				if( _areLanguageNotificationsActivated )
				{
					Iterator<ChangeLanguageClientInterface> it = _listOfObserversOfLanguageChanges.iterator();
					while( it.hasNext() )
					{
						it.next().changeLanguage( language );
					}
				}

				_avoidRecursion = true;
				setLanguage( language );
			}
			catch( Throwable th )
			{
				th.printStackTrace();
				if( revertOnError )
				{
					_avoidRecursion = false;
					changeLanguage_internal( oldLanguage, false );
					throw( new ConfigurationException( "Could not change language, because of an internal error: " + th.getMessage() ) );
				}
			}
			finally
			{
				_avoidRecursion = false;
			}
		}
	}

	public void unregisterFromChangeLanguageAsObserver()
	{
		if( _changeLanguageServer != null )
		{
			_changeLanguageServer.unregisterChangeLanguageObserver( this );
		}
	}

	public void registerToChangeLanguageAsObserver( ChangeLanguageServerInterface conf)
	{
		if( ( _changeLanguageServer == null ) && ( conf != null ) )
		{
			_changeLanguageServer = conf;
			_changeLanguageServer.registerChangeLanguageObserver( this );
		}
	}

	public void activateChangeLanguageNotifications( boolean value )
	{
		_areLanguageNotificationsActivated = value;
	}

	public boolean areChangeLanguageNotificationsActivated()
	{
		return( _areLanguageNotificationsActivated );
	}
	
	@Override
	public void registerChangeLanguageObserver( ChangeLanguageClientInterface requestor )
	{
		int index = CollectionFunctions.indexOfReference(_listOfObserversOfLanguageChanges, requestor);

		if( index < 0 )
		{
			_listOfObserversOfLanguageChanges.add( requestor );
		}
	}

	@Override
	public void unregisterChangeLanguageObserver( ChangeLanguageClientInterface requestor )
	{
		int index = CollectionFunctions.indexOfReference(_listOfObserversOfLanguageChanges, requestor);

		if( index >= 0 )
		{
			_listOfObserversOfLanguageChanges.remove( requestor );
		}
	}

	public String M_getConfigurationFileName()
	{
		return( M_getConfigurationFileName( a_language ) );
	}

	protected String M_getConfigurationFileName( String language )
	{
		String fileName = a_rootApplicationConfigurationPath;
		if( language != null ) fileName = fileName + sa_dirSeparator + language;
		fileName = fileName + sa_dirSeparator + a_configurationFileName;

		return( fileName );
	}

	protected Properties M_loadConfigurationFile( String fileName )
	{
		Properties result = null;
		try
		{
			a_existedConfigurationFile = false;
			result = M_loadProperties( fileName );
			a_existedConfigurationFile = true;
		}
		catch( IOException ioe )
		{
//			ioe.printStackTrace();  // we do not retrhrow the exception, and we work with default values.
			System.out.println( ioe.getMessage() );
		}

		return( result );
	}

	protected Properties M_loadProperties( String filename ) throws IOException
	{
		File file = new File( filename );
		Properties result = new Properties();

		FileInputStream fis = null;
		InputStreamReader isr = null;

		try
		{
			fis = new FileInputStream(file);
			isr = new InputStreamReader( fis, StandardCharsets.UTF_8 );
			result.load( isr );
		}
		catch( Throwable th )
		{
			System.out.println( "Error reading properties file: " + th.getMessage() );
			throw( th );
		}
		finally
		{
			if( isr != null ) isr.close();
			else if( fis != null )	fis.close();
		}

		return( result );
	}

	public int M_getIntParamConfiguration( String label )
	{
		return( (int) M_getLongParamConfiguration( label ) );
	}

	public long M_getLongParamConfiguration( String label )
	{
		long result = -1;
		String resultStr = M_getStrParamConfiguration( label );
		try
		{
			result = Long.parseLong( resultStr );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			System.out.println( "Error when reading numerical configuration parameter: " +
								label );
			result = -1;
		}

		return( result );
	}

	public boolean M_getBoolParamConfiguration( String label )
	{
		boolean result = false;
		String resultStr = M_getStrParamConfiguration( label );

		if( resultStr != null )
		{
			result = !resultStr.equals("0");
		}

		return( result );
	}

	public String M_getStrParamConfiguration( String label )
	{
		String result = null;
		if( a_properties != null ) result = a_properties.getProperty(label);
		if( (result == null) && (a_defaultProperties != null) ) result = a_defaultProperties.getProperty(label);
		if( (result == null) && (a_defaultDefaultProperties != null) ) result = a_defaultDefaultProperties.getProperty(label);
		return( result );
	}

	public float M_getFloatFactor( String label )
	{
		float factor = -1.0F;
		String resultStr = M_getStrParamConfiguration( label );

		try
		{
			factor = Float.valueOf( resultStr );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
		return( factor );
	}
	
	public double M_getDoubleParamConfiguration( String label )
	{
		double result = -1.0D;
		String resultStr = M_getStrParamConfiguration( label );

		try
		{
			result = Double.valueOf( resultStr );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
		return( result );
	}
	
	
	public void M_setIntParamConfiguration( String label, int value )
	{
		M_setStrParamConfiguration( label, String.valueOf( value ) );
	}
	
	public void M_setLongParamConfiguration( String label, long value )
	{
		M_setStrParamConfiguration( label, String.valueOf( value ) );
	}

	public void M_setBoolParamConfiguration( String label, boolean value )
	{
		M_setStrParamConfiguration( label, ( value ? "1" : "0" ) );
	}

	public void M_setStrParamConfiguration( String label, String value )
	{
		if( a_properties == null ) a_properties = new Properties();
		a_properties.setProperty( label, value );
	}

	public void M_setFloatParamConfiguration( String label, float value )
	{
		if( a_properties == null ) a_properties = new Properties();
		a_properties.setProperty( label, String.valueOf( value ) );
	}

	public void M_setDoubleParamConfiguration( String label, double value )
	{
		if( a_properties == null ) a_properties = new Properties();
		a_properties.setProperty( label, String.valueOf( value ) );
	}

	protected Properties M_makePropertiesAddingDefaults( Properties values, Properties defaults )
	{
		Properties result = null;
		
		if( defaults == null ) result = values;
		else if( values == null ) result = defaults;
		else
		{
			result = (Properties) values.clone();
			Object[] labels = defaults.stringPropertyNames().toArray();
			for( int ii=0; ii<labels.length; ii++ )
			{
				String key = (String) labels[ii];
				String value = result.getProperty(key );
				if( value == null )
				{
					value = defaults.getProperty( key );
					if( value != null ) result.setProperty( key, value );
				}
			}
		}
		
		return( result );
	}
	
	public void M_saveConfiguration() throws ConfigurationException
	{
		M_saveConfiguration( false );
	}

	public void M_saveConfiguration( boolean saveAlways ) throws ConfigurationException
	{
		if( ! saveAlways )
			M_saveConfiguration_notAlways( M_getConfigurationFileName(a_language) );
		else
			M_saveConfiguration( M_getConfigurationFileName(a_language) );
	}

	protected void M_saveConfiguration_notAlways( String fileName ) throws ConfigurationException
	{
		File file = new File( fileName );

		// we only save the language properties file once
		if( ( a_language == null ) || ( !file.isFile() ) )
			M_saveConfiguration( fileName );
	}

	public void M_saveConfiguration( String fileName ) throws ConfigurationException
	{
		File file = new File( fileName );
		File fPath = new File( file.getParent() );

		if( !fPath.exists() )
		{
			fPath.mkdirs();
		}
		else if ( !fPath.isDirectory() )
		{
			throw( new ConfigurationException(  "Error. Could not save the " +
												"configuration because " + fPath.getName() +
												" is not a directory"  ) );
		}

		Properties prop = M_makePropertiesAddingDefaults( a_properties, a_defaultProperties );
		prop = M_makePropertiesAddingDefaults( prop, a_defaultDefaultProperties );

		if( prop != null )
		{
			FileOutputStream fos = null;
			OutputStreamWriter osw = null;
			try
			{
				fos = new FileOutputStream( file );
				osw = new OutputStreamWriter( fos, StandardCharsets.UTF_8 );
				prop.store( osw, a_configurationFileName + " Configuration" );
			}
			catch( IOException ex )
			{
				ex.printStackTrace();
				throw( new ConfigurationException(  "Error. Could not save the configuration file " +
													"because of " + ex.getMessage() ) );
			}
			finally
			{
				try
				{
					if( osw != null ) osw.close();
					else if( fos != null ) fos.close();
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
			}
		}
	}

	public String M_getLanguage()	{ return( a_language );	}

	@Override
	public String getProperty( String label )
	{
		return( M_getStrParamConfiguration( label ) );
	}

	@Override
	public Object setProperty( String key, String value )
	{
		M_setStrParamConfiguration( key, value );
		return( null );
	}

	protected Properties cargarPropertiesClassPath( String propName ) throws IOException
	{
		a_existedConfigurationFileInClassPath = false;
		Properties result = null;
		InputStream in = null;
		ClassLoader loader = ClassLoader.getSystemClassLoader ();
		in = loader.getResourceAsStream (propName);

		if( in == null )
		{
			try
			{
				in = this.getClass().getClassLoader().getResource(propName).openStream();
			}
			catch( Throwable th )
			{
				in = null;
			}
		}

		if (in != null)
		{
			InputStreamReader isr = new InputStreamReader( in, StandardCharsets.UTF_8 );
			if( isr != null )
			{
				result = new Properties ();
				result.load (isr); // It can throw IOException
				isr.close();
			}
		}
		a_existedConfigurationFileInClassPath = ( result != null );
		return( result );
	}

	public boolean M_isFirstTime()
	{
		return( !a_existedConfigurationFile );
	}

	public boolean M_existedConfigurationFileInClassPath()
	{
		return( a_existedConfigurationFileInClassPath );
	}

	public void M_setProperties( Properties prop )
	{
		a_properties = prop;
	}

	public void loadDefaultProperties()
	{
		a_defaultProperties=M_getDefaultProperties( a_language );
	}

	public void registerConfigurationLabel( String label, String defaultValue )
	{
		a_defaultDefaultProperties.setProperty( label, defaultValue );
	}

	public static void main( String[] args )
	{
		String fileName = "C:\\Users\\Usuario\\frojasg1.apps\\FileEncoderApplication\\Frames\\CAT\\JDial_applicationConfiguration_LAN.properties";
		
		File file = new File( fileName );
		Properties result = new Properties();

		try
		{

			FileInputStream fis = new FileInputStream(file);
			InputStreamReader isr = new InputStreamReader( fis, StandardCharsets.UTF_8 );
			result.load( isr );
			isr.close();

			boolean success = file.delete();

			System.out.println( "Delete " + file + "   Success: " + success );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}
}
